import type { GraphQueryResultFromBackend, GraphStatistics } from 'ts-common';
import { getAttributeType, initializeStatistics, updateStatistics } from './utils';

export const getGraphStatistics = (graph: GraphQueryResultFromBackend): GraphStatistics => {
  const { nodes, edges } = graph;

  const n_nodes = nodes.length;
  const n_edges = edges.length;

  const density = n_nodes < 2 ? 0 : (n_edges * 2) / (n_nodes * (n_nodes - 1));

  // general nodes and edges statistics
  const metaData: GraphStatistics = {
    topological: { density, self_loops: 0 },
    nodes: { labels: [], count: n_nodes, types: {} },
    edges: { labels: [], count: n_edges, types: {} },
  };
  // return metaData;

  // attributes based statistics
  // Pre-initialize maps for faster lookups
  const seenNodeTypes = new Set();

  for (let i = 0; i < nodes.length; i++) {
    const node = nodes[i];
    const nodeType = node.label;

    // Use Set for faster existence checks
    if (!seenNodeTypes.has(nodeType)) {
      seenNodeTypes.add(nodeType);
      metaData.nodes.labels.push(nodeType);
      metaData.nodes.types[nodeType] = { count: 0, attributes: {} };
    }

    const nodeTypeStats = metaData.nodes.types[nodeType];
    nodeTypeStats.count++;

    const attributes = node.attributes;
    const nodeTypeAttributes = nodeTypeStats.attributes;

    const attributeEntries = Object.entries(attributes);
    for (let j = 0; j < attributeEntries.length; j++) {
      const [attributeId, attributeValue] = attributeEntries[j];
      if (!nodeTypeAttributes[attributeId]) {
        const attributeType = getAttributeType(attributeValue);
        nodeTypeAttributes[attributeId] = {
          attributeType,
          statistics: initializeStatistics(attributeType)
        };
      }
      updateStatistics(nodeTypeAttributes[attributeId], attributeValue);
    }
  };

  // Process edges
  // Pre-initialize sets and maps for faster lookups
  const seenEdgeTypes = new Set();

  for (let i = 0; i < edges.length; i++) {
    const edge = edges[i];
    const edgeType = edge.label;

    // Use Set for faster existence checks
    if (!seenEdgeTypes.has(edgeType)) {
      seenEdgeTypes.add(edgeType);
      metaData.edges.labels.push(edgeType);
      metaData.edges.types[edgeType] = { count: 0, attributes: {} };
    }

    metaData.edges.types[edgeType].count++;

    if (edge.from === edge.to) metaData.topological.self_loops++;

    const edgeTypeAttributes = metaData.edges.types[edgeType].attributes;
    const attributes = edge.attributes;

    const attributeEntries = Object.entries(attributes);
    for (let i = 0; i < attributeEntries.length; i++) {
      const [attributeId, attributeValue] = attributeEntries[i];
      if (!edgeTypeAttributes[attributeId]) {
        const attributeType = getAttributeType(attributeValue);
        edgeTypeAttributes[attributeId] = {
          attributeType,
          statistics: initializeStatistics(attributeType)
        };
      }
      updateStatistics(edgeTypeAttributes[attributeId], attributeValue);
    }
  };

  return metaData;
};

