import {
  getUserSaveState,
  getUserSaveStateInsight,
  type BackendQueryFormat,
  type QueryRequest,
  type GraphQueryResultFromBackend,
  type GraphQueryResultMetaFromBackend,
  mlDefaultState,
  updateUserSaveStateInsight,
} from "ts-common";
import { Neo4jConnection, type DbConnection, RabbitMqBroker, RedisConnector } from "ts-common";
import {
  RABBIT_HOST,
  RABBIT_PASSWORD,
  RABBIT_PORT,
  RABBIT_USER,
  USER_MANAGEMENT_SERVICE_API,
  type QueryExecutionTypes,
} from "../variables";
import { log } from "../logger";
import { QueryPublisher } from "../utils/queryPublisher";
import { query2Cypher } from "../utils/cypher/converter";
import { parseCypherQuery } from "../utils/cypher/queryParser";
import { formatTimeDifference } from "ts-common/src/logger/logger";
import { graphQueryBackend2graphQuery } from "../frontend/statistics";
import { Query2BackendQuery } from "../utils/reactflow/query2backend";
import { hashDictionary, hashIsEqual } from "../utils/hashing";

export const queryService = async (db: DbConnection, query: string): Promise<GraphQueryResultMetaFromBackend> => {
  // TODO: only neo4j is supported for now
  const connection = new Neo4jConnection(db);
  try {
    const [neo4jResult] = await connection.run([query]);
    const graph = parseCypherQuery(neo4jResult.records);

    // calculate metadata
    const result = graphQueryBackend2graphQuery(graph);

    return result;
  } catch (error) {
    log.error("Error parsing query result:", query, error);
    throw new Error("Error parsing query result");
  } finally {
    connection.close();
  }
};

export const queryServiceReader = async (
  frontendPublisher: RabbitMqBroker,
  mlPublisher: RabbitMqBroker,
  redis: RedisConnector,
  type: QueryExecutionTypes
) => {
  if (type == null) {
    log.error("Unsupported query execution type:", type);
    throw new Error("Unsupported query execution type");
  }
  log.info("Starting query reader for", type);

  const publisher = new QueryPublisher(frontendPublisher, mlPublisher);

  const queryServiceConsumer = new RabbitMqBroker("requests-exchange", `${type}-query-queue`, `${type}-query-request`);
  await queryServiceConsumer.connect({
    protocol: "amqp",
    hostname: RABBIT_HOST,
    port: RABBIT_PORT,
    username: RABBIT_USER,
    password: RABBIT_PASSWORD,
  });

  log.info("Connected to RabbitMQ ST!");

  await queryServiceConsumer.startConsuming<QueryRequest>("query-service", async (message, headers) => {
    const startTime = Date.now();
    const ss = await getUserSaveState(
      headers.message.sessionData.userID,
      headers.message.sessionData.saveStateID,
      USER_MANAGEMENT_SERVICE_API
    );
    const routingKey = await redis.getRouting(headers.message.sessionData.sessionID);

    log.debug("Received query request:", message, headers, ss);
    log.debug("Received routing key:", routingKey);

    if (!headers.queryID) {
      log.error("QueryID not set in headers:", headers);
      return;
    }

    log.debug("Received query request:", message, headers);

    publisher.withHeaders(headers).withRoutingKey(routingKey.QueueID).withQueryID(headers.queryID);
    publisher.publishStatusToFrontend("Received");

    if (ss == null || ss.dbConnections == null || ss.dbConnections[0] == null || ss.dbConnections.length === 0) {
      log.error("Invalid SaveState received in queryServiceConsumer:", ss);
      publisher.publishErrorToFrontend("Invalid SaveState");
      return;
    }

    const visualQuery = ss.queries[0].graph;
    const queryBuilderSettings = ss.queries[0].settings;
    const ml = message.ml;
    const convertedQuery = Query2BackendQuery(ss.id, visualQuery, queryBuilderSettings, ml);

    log.debug("translating query:", convertedQuery);
    publisher.publishStatusToFrontend("Translating");

    const query = query2Cypher(convertedQuery);
    if (query == null) {
      log.error("Error translating query:", convertedQuery);
      publisher.publishErrorToFrontend("Error translating query");
      return;
    }

    log.debug("Translated query:", query);
    publisher.publishTranslationResultToFrontend(query);

    for (let i = 0; i < ss.dbConnections.length; i++) {
      queryService(ss.dbConnections[i], query)
        .then((result) => {
          publisher.publishResultToFrontend(result);
          log.debug("Query result!");
          log.info(`Query executed in ${formatTimeDifference(Date.now() - startTime)}`);

          if (convertedQuery.machineLearning && convertedQuery.machineLearning.length > 0) {
            for (let i = 0; i < convertedQuery.machineLearning.length; i++) {
              try {
                publisher.publishMachineLearningRequest(result, convertedQuery.machineLearning[i], headers);
                log.debug("Published machine learning request", convertedQuery.machineLearning[i]);
              } catch (error) {
                log.error("Error publishing machine learning request", error);
                publisher.publishErrorToFrontend("Error publishing machine learning request");
              }
            }
          }
        })
        .catch((error) => {
          log.error("Error querying database", error);
          publisher.publishErrorToFrontend("Error querying database");
        });
    }
  });
};

export const queryServiceReaderStatsCheck = async (type: QueryExecutionTypes) => {
  if (type == null) {
    log.error("Unsupported query execution type:", type);
    throw new Error("Unsupported query execution type");
  }

  // TODO: change requests-exchange -> requests-exchange-stats
  const queryServiceConsumer = new RabbitMqBroker("requests-exchange", `${type}-query-queue`, `${type}-query-request`);

  await queryServiceConsumer.connect({
    protocol: "amqp",
    hostname: RABBIT_HOST,
    port: RABBIT_PORT,
    username: RABBIT_USER,
    password: RABBIT_PASSWORD,
  });

  log.info("Connected to RabbitMQ ST!");

  // query-service to query-service-insight-reader
  await queryServiceConsumer.startConsuming<QueryRequest>("query-service-insight-reader", async (message, headers) => {
    const startTime = Date.now();
    const ss = await getUserSaveState(
      headers.message.sessionData.userID,
      headers.message.sessionData.saveStateID,
      USER_MANAGEMENT_SERVICE_API
    );
    // !TODO: send from frontend the condition operator?
    type InsightCondition = "Greater than" | "Equal than" | "Smaller than";
    /*
    const ssInsight = await getUserSaveStateInsight(
      headers.message.sessionData.userID,
      headers.message.sessionData.saveStateID,
      USER_MANAGEMENT_SERVICE_API
    );
    
    const ssInsightNode = ssInsight.statsCheck.node;
    const ssInsightStatistic = ssInsight.statsCheck.statistic;
    const ssInsightCondition = ssInsight.statsCheck.condition; as InsightCondition;
    const ssInsightValue = ssInsight.statsCheck.value;
    */

    const ssInsightNode = "Movie";
    const ssInsightStatistic = "Count";
    const ssInsightCondition = "Equal than" as InsightCondition;
    const ssInsightValue = 10;

    log.debug("Received query request:", ss);

    if (!headers.queryID) {
      log.error("QueryID not set in headers:", headers);
      return;
    }

    log.debug("Received check stat request:", message, headers);

    if (ss == null || ss.dbConnections == null || ss.dbConnections[0] == null || ss.dbConnections.length === 0) {
      log.error("Invalid SaveState received in queryServiceConsumer:", ss);
      return;
    }

    const visualQuery = ss.queries[0].graph;
    const queryBuilderSettings = ss.queries[0].settings;
    const ml = message.ml;
    const convertedQuery = Query2BackendQuery(ss.id, visualQuery, queryBuilderSettings, ml);

    log.debug("translating query:", convertedQuery);

    const query = query2Cypher(convertedQuery);
    if (query == null) {
      log.error("Error translating query:", convertedQuery);
      return;
    }

    for (let i = 0; i < ss.dbConnections.length; i++) {
      queryService(ss.dbConnections[i], query)
        .then((result) => {
          log.debug("Query result!");
          log.info(`Query executed in ${formatTimeDifference(Date.now() - startTime)}`);

          if (result.metaData.nodes.labels.includes(ssInsightNode)) {
            // TODO: Check if other stats are available
            const nodeCount = result.metaData.nodes.count;

            let conditionMet = false;
            switch (ssInsightCondition) {
              case "Greater than":
                conditionMet = nodeCount > ssInsightValue;
                break;
              case "Equal than":
                conditionMet = nodeCount === ssInsightValue;
                break;
              case "Smaller than":
                conditionMet = nodeCount < ssInsightValue;
                break;
              default:
                log.error(`Unsupported condition: ${ssInsightCondition}`);
                throw new Error(`Unsupported condition: ${ssInsightCondition}`);
            }
            log.info(`Nodes: ${nodeCount}. Value: ${ssInsightValue}. Condition: ${ssInsightCondition}. Met: ${conditionMet}`);

            if (conditionMet) {
              log.info("Condition met ");

              // TODO: Publish message accordingly
            } else {
              log.info("Condition not met");
            }
          }
        })
        .catch((error) => {
          log.error("Error querying database", error);
        });
    }
  });
};

export const queryServiceReaderDiffCheck = async (type: QueryExecutionTypes) => {
  if (type == null) {
    log.error("Unsupported query execution type:", type);
    throw new Error("Unsupported query execution type");
  }

  // TODO: change requests-exchange -> requests-exchange-stats
  const queryServiceConsumer = new RabbitMqBroker("requests-exchange", `${type}-query-queue`, `${type}-query-request`);

  await queryServiceConsumer.connect({
    protocol: "amqp",
    hostname: RABBIT_HOST,
    port: RABBIT_PORT,
    username: RABBIT_USER,
    password: RABBIT_PASSWORD,
  });

  log.info("Connected to RabbitMQ ST!");

  // query-service to query-service-insight-reader
  await queryServiceConsumer.startConsuming<QueryRequest>("query-service-insight-reader", async (message, headers) => {
    const startTime = Date.now();
    const ss = await getUserSaveState(
      headers.message.sessionData.userID,
      headers.message.sessionData.saveStateID,
      USER_MANAGEMENT_SERVICE_API
    );

    const ssInsight = await getUserSaveStateInsight(
      headers.message.sessionData.userID,
      headers.message.sessionData.saveStateID,
      USER_MANAGEMENT_SERVICE_API
    );
    // const previousQueryResult = ssInsight.previous_result_hash;
    const previousQueryResult = "Get hash from db";

    log.debug("Received query request:", ss);

    if (!headers.queryID) {
      log.error("QueryID not set in headers:", headers);
      return;
    }

    log.debug("Received check stat request:", message, headers);

    if (ss == null || ss.dbConnections == null || ss.dbConnections[0] == null || ss.dbConnections.length === 0) {
      log.error("Invalid SaveState received in queryServiceConsumer:", ss);
      return;
    }

    const visualQuery = ss.queries[0].graph;
    const queryBuilderSettings = ss.queries[0].settings;
    const ml = message.ml;
    const convertedQuery = Query2BackendQuery(ss.id, visualQuery, queryBuilderSettings, ml);

    log.debug("translating query:", convertedQuery);

    const query = query2Cypher(convertedQuery);
    if (query == null) {
      log.error("Error translating query:", convertedQuery);
      return;
    }

    for (let i = 0; i < ss.dbConnections.length; i++) {
      queryService(ss.dbConnections[i], query)
        .then((result) => {
          log.debug("Query result!");
          log.info(`Query executed in ${formatTimeDifference(Date.now() - startTime)}`);

          const queryResult = hashDictionary({
            nodes: result.nodes.map((node) => node._id),
            edges: result.edges.map((edge) => edge._id),
          });

          log.info("Comparing hash values from current and previous query");

          if (previousQueryResult && !hashIsEqual(queryResult, previousQueryResult)) {
            log.info("Different results, use Dennis code...");
          } else {
            log.info("No difference in result sets");
          }

          // Save query result for future reference
          updateUserSaveStateInsight(
            headers.message.sessionData.userID,
            headers.message.sessionData.saveStateID,
            USER_MANAGEMENT_SERVICE_API,
            { previous_result_hash: queryResult }
          );

          log.info("Updated node and edge ids in SaveState");
        })
        .catch((error) => {
          log.error("Error querying database", error);
        });
    }
  });
};
